#include "SceneLoader.h"

#include "src/scene/Scene.h"

#include "src/scene/surfaces/Checkerboard.h"
#include "src/scene/surfaces/Quadric.h"
#include "src/scene/surfaces/Sphere.h"


#include <fstream>
#include <memory>

using json = nlohmann::json;

Scene SceneLoader::loadScene(std::string path){
    using namespace nlohmann;

    Scene scene;

    std::ifstream file(path);
    json data;
    try {
        data = json::parse(file);

        Vec4f cameraTarget(data["camera"]["target"].get<std::vector<float>>());
        float cameraDistance = data["camera"]["distance"].get<float>();
        float focalLength = data["camera"]["focalLength"].get<float>();

        float xRotation = data["camera"].contains("xRotation") ? data["camera"]["xRotation"].get<float>() : 0;
        float yRotation = data["camera"].contains("yRotation") ? data["camera"]["yRotation"].get<float>() : 0;

        scene.camera = Camera(cameraTarget, cameraDistance, focalLength, xRotation, yRotation);

        std::vector<Shader> shaders;
        // Materials
        for (const auto& item : data["materials"].items())
        {
            Shader shader;
            shader.ambientColor = Vec4f(item.value()["ambient"].get<std::vector<float>>());
            shader.diffuseColor = Vec4f(item.value()["diffuse"].get<std::vector<float>>());
            shader.reflectiveColor = Vec4f(item.value()["reflective"].get<std::vector<float>>());
            shader.refractiveColor = Vec4f(item.value()["refractive"].get<std::vector<float>>());
            shader.refractiveIndex = item.value()["refractiveIndex"].get<float>();
            shader.shininess = item.value()["shininess"].get<float>();
            shaders.push_back(shader);
        }

        // Surfaces
        for (const auto& item : data["surfaces"].items())
        {
            if(item.value()["type"] == "Sphere"){
                scene.surface.addSurface(std::make_shared<Sphere>(
                    Vec4f(item.value()["center"].get<std::vector<float>>()),
                    item.value()["radius"].get<float>(),
                    shaders[item.value()["material"].get<int>()]
                    ));
            } else if(item.value()["type"] == "Checkerboard"){
                scene.surface.addSurface(std::make_shared<Checkerboard>(
                    Vec4f(item.value()["position"].get<std::vector<float>>()),
                    Vec4f(item.value()["x"].get<std::vector<float>>()),
                    Vec4f(item.value()["y"].get<std::vector<float>>()),
                    shaders[item.value()["material"].get<int>()],
                    shaders[item.value()["material2"].get<int>()]
                ));
            } else if(item.value()["type"] == "Quadric"){
                scene.surface.addSurface(std::make_shared<Quadric>(
                    Mat4f(item.value()["Q"].get<std::vector<float>>()),
                    shaders[item.value()["material"].get<int>()]
                ));
            }
        }

        // Lights:
        std::vector<PointLight> lights;
        for (const auto& item : data["lights"].items())
        {
            if(item.value()["type"] == "PointLight"){
                lights.push_back(PointLight(
                    Vec4f(item.value()["position"].get<std::vector<float>>()),
                    Vec4f(item.value()["color"].get<std::vector<float>>())
                ));
            }
        }
        scene.lights = lights;
    } catch(json::exception& ex){
        std::cout << ex.what() << std::endl;
    }

    return scene;
}
